package moj.lang.statserv;

import java.io.Reader;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A StatServRequest module that requests and supplies frequencies on PoS 
 * tagged words.
 *
 * @author  Martin Hassel
 * @version 2006-Dec-13
 */
public class WordTagRequest extends StatServRequest {
	private static final String moduleName = "andrist.statservmod.wordtag.WordTagModule"; 
	// For storing statistics
	private Map<String,Integer> localTF = null;
	private Map<String,Integer> referenceTF = null;
	private Map<String,Integer> referenceDF = null;
	// Text and doc.set tf/df stats
	public int localTokens = 0;
	public int referenceTokens = 0;
	public int referenceDocuments = 0;

	/**
	 * Creates new <code>WordTagRequest</code> and fills mandatory fields.
	 * @param docset the document set statistics is requested for
	 * @param localstats is local statistics requested?
	 * @param referencestats is reference statistics requested?
	 * @param addstats is statistics to be added to the document set?
	 */
	public WordTagRequest(String docset, boolean localstats, boolean referencestats, boolean addstats) {
		super(moduleName, docset, localstats, referencestats, addstats);
	}

	/**
	 * Returns the XML that constitutes the modules part in a 
	 * <code>StatServConnection</code> request.  
	 * @return XML representing the modules part in a request
	 */
	public String requestXML() {
		return "<module>\n<class-name>" + moduleName + "</class-name>\n</module>\n";
	}

	/**
	 * Parses the XML passed via the <code>Reader</code> and fills the modules 
	 * data structures. 
	 * @param xmlReader <code>Reader</code> wrapping the XML response from a 
	 * <code>StatServConnection</code> request.
	 * @return <code>true</code> on success, else <code>false</code>
	 */
	public boolean parse(Reader xmlReader) {
		try {
			SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
			parser.parse(new InputSource(xmlReader), new WordTagHandler());
			return true;
		} catch (Exception ex) {
			//ex.printStackTrace();
			return false;
		}
	}

	/**
	 * @return Returns the localTF.
	 */
	public Map<String,Integer> getLocalTF() {
		return localTF;
	}

	/**
	 * @return Returns the number of tokens in the reference data set.
	 */
	public int getLocalTokens() {
		return localTokens;
	}

	/**
	 * @return Returns the number of documents in the reference data set.
	 */
	public int getReferenceDocuments() {
		return referenceDocuments;
	}

	/**
	 * @return Returns the number of tokens in the reference data set.
	 */
	public int getReferenceTokens() {
		return referenceTokens;
	}

	/**
	 * @return Returns the referenceDF.
	 */
	public Map<String,Integer> getReferenceDF() {
		return referenceDF;
	}

	/**
	 * @return Returns the referenceTF.
	 */
	public Map<String,Integer> getReferenceTF() {
		return referenceTF;
	}

	private class WordTagHandler extends DefaultHandler {
		private boolean inModule = false;
		private boolean inLocStats = false;
		private boolean inRefStats = false;
		private boolean inWord = false;
		private StringBuffer currentWord = null;
		private String currentTag = null;
		private int currentTF = 0;
		private int currentDF = 0;
		
		@SuppressWarnings("synthetic-access")
		public void startElement(String namespaceURI, String localName,
				String qualifiedName, Attributes attributes) {
			// <module ...>
			if(qualifiedName.equals("module"))
				if(attributes.getValue("name").equals(moduleName))
					inModule = true;
			// <reference-stats ...>
			if(qualifiedName.equals("local-stats") && inModule == true) {
				localTokens = Integer.parseInt(attributes.getValue("words"));
				localTF = new HashMap<String,Integer>();
				inLocStats = true;
			}
			// <reference-stats ...>
			if(qualifiedName.equals("reference-stats") && inModule == true) {
				referenceDocuments = Integer.parseInt(attributes.getValue("docs"));
				referenceTokens = Integer.parseInt(attributes.getValue("words"));
				referenceTF = new HashMap<String,Integer>();
				referenceDF = new HashMap<String,Integer>();
				inRefStats = true;
			}
			// <word ...>
			if(qualifiedName.equals("word") && inModule == true && inLocStats == true) {
				currentTag = attributes.getValue("tag");
				currentTF = Integer.parseInt(attributes.getValue("tf"));
				currentWord = new StringBuffer();
				inWord = true;
			}
			// <word ...>
			if(qualifiedName.equals("word") && inModule == true && inRefStats == true) {
				currentTag = attributes.getValue("tag");
				currentTF = Integer.parseInt(attributes.getValue("tf"));
				currentDF = Integer.parseInt(attributes.getValue("df"));
				currentWord = new StringBuffer();
				inWord = true;
			}
		}
		
		@SuppressWarnings("synthetic-access")
		public void endElement(String namespaceURI, String localName,
				String qualifiedName) {
			// </word>
			if(qualifiedName.equals("word") && inModule == true && inLocStats == true) {
				localTF.put((currentWord+"_").replaceAll(" ","_") + currentTag, currentTF);
				currentWord = null;
				currentTag = null;
				currentTF = 0;
				inWord = false;
			}
			// </word>
			if(qualifiedName.equals("word") && inModule == true && inRefStats == true) {
				referenceTF.put((currentWord+"_").replaceAll(" ","_") + currentTag, currentTF);
				referenceDF.put((currentWord+"_").replaceAll(" ","_") + currentTag, currentDF);
				currentWord = null;
				currentTag = null;
				currentTF = 0;
				currentDF = 0;
				inWord = false;
			}
			// </reference-stats>
			if(qualifiedName.equals("local-stats"))
				inLocStats = false;
			// </reference-stats>
			if(qualifiedName.equals("reference-stats"))
				inRefStats = false;
			// </module>
			if(qualifiedName.equals("module"))
				inModule = false;
		}
		
		public void characters(char[] ch, int start, int length) {
			if(inWord)
				if(currentWord != null)
					currentWord.append(ch, start, length);
		}
	} // End nested class
}
