/*
 * SparseDistributedMemory.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is part of SweSum^2 (see http://swesum.nada.kth.se/),
 * and is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2006-Oct-17
 * http://www.nada.kth.se/~xmartin/
 *
 */

package moj.ri;

import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import moj.util.TagStripper;

import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

/**
 * @author  Martin Hassel
 * @version 2006-Dec-07
 */
public class IndexHandler extends DefaultHandler {
    // These are per RandomIndex, used as 'return values' and therefore public
    public ConcurrentHashMap<String,RandomLabel> randomIndex = new ConcurrentHashMap<String,RandomLabel>();
	public Set<String> vocabulary = new HashSet<String>();
    public int dimensionality = 0;
    public int randomDegree = 0;
    public int randomSeed = 0;
    public int leftWindowSize = 0;
    public int rightWindowSize = 0;
    public String weightingScheme = "";
    public long wordsIndexed = 0;
    public long documentsIndexed = 0;
    public boolean allLowercase = true;
    public boolean unaryLabels = false;
    public boolean documentLabels = false;
    public boolean createAhead = true;
    public boolean purged = false;
    // These are per RandomLabel read from the XML
    private String _word;	// The word the 'label' is to be attached to
    private String _meta_data;	 // Field available for meta data (e.g. PoS?)
    private long _termfrequency; // Number of updates done to the label
    private int  _docfrequency;	 // Number of documents the word occurs in
	private int _negs[];	// Posititions for -1:s in the random label
	private int _poss[];	// Posititions for 1:s in the random label
    private float _context[];	// The actual contextually updated label
    // Buffer to read chunk data into
    private StringBuffer buffer = null;
    private boolean progressIndicator = false;
	private boolean _inVocabulary = false;
	private boolean _foundCV = false;
	private int no_pos_and_neg;

    public void startElement(String namespaceURI, String localName,
            String qualifiedName, Attributes attributes) {
        buffer = new StringBuffer();

        // <randomlabels>
        if(qualifiedName.equals("randomlabels")) {
            dimensionality = Integer.parseInt(attributes.getValue("dimensionality"));
            randomDegree = Integer.parseInt(attributes.getValue("randomness"));
            randomSeed = Integer.parseInt(attributes.getValue("seed"));
            leftWindowSize = Integer.parseInt(attributes.getValue("leftWindowSize"));
            rightWindowSize = Integer.parseInt(attributes.getValue("rightWindowSize"));
            weightingScheme = attributes.getValue("weightingScheme");
            documentsIndexed = Long.parseLong(attributes.getValue("documentsIndexed"));
            if(attributes.getValue("allLowercase")!=null)
                allLowercase = (attributes.getValue("allLowercase").compareToIgnoreCase("true")==0) ? true : false;
            if(attributes.getValue("unaryLabels")!=null)
                unaryLabels = (attributes.getValue("unaryLabels").compareToIgnoreCase("true")==0) ? true : false;
            if(attributes.getValue("documentLabels")!=null)
                documentLabels = (attributes.getValue("documentLabels").compareToIgnoreCase("true")==0) ? true : false;
            if(attributes.getValue("createAhead")!=null)
                createAhead = (attributes.getValue("createAhead").compareToIgnoreCase("true")==0) ? true : false;
            if(attributes.getValue("purged")!=null)
                purged = (attributes.getValue("purged").compareToIgnoreCase("true")==0) ? true : false;
            // Unused so far but enhances the informativeness of the XML to human eyes
            // Could be used for content validating beyond well-formedness
            // Long.parseLong(atts.getValue("wordsIndexed"));
            // Long.parseLong(atts.getValue("labels"));
            no_pos_and_neg = Math.round(randomDegree/2);
        }

        // <rl>
        if(qualifiedName.equals("rl")) {
            _word = "";
            _meta_data = null;
            _termfrequency = 0;
            _docfrequency = 0;
    		_negs = new int[no_pos_and_neg];
    		_poss = new int[no_pos_and_neg];
            _context = new float[dimensionality];
        }

        // <vocabulary>
        if(qualifiedName.equals("vocabulary"))
			_inVocabulary = true;
	}

    public void endElement(String namespaceURI, String localName, String qualifiedName) {
        String chunk = buffer.toString();

        // </vocabulary>
        if(qualifiedName.equals("vocabulary"))
			_inVocabulary = false;

        // </wd>
        if(qualifiedName.equals("wd")) {
            _word = TagStripper.fromXMLentities(chunk);
			if(_inVocabulary)
				vocabulary.add(_word);
        }

        // </md>
        if(qualifiedName.equals("md"))
            _meta_data = chunk.length()>0 ? TagStripper.fromXMLentities(chunk) : null;

        // </tf>
        if(qualifiedName.equals("tf"))
            _termfrequency = Long.parseLong(chunk);

        // </df>
        if(qualifiedName.equals("df"))
            _docfrequency  = Integer.parseInt(chunk);

        // </neg>
        if(qualifiedName.equals("neg")) {
        	if(chunk.length() > 0) {
                String data_array[] = chunk.split(",");
                for(int i = 0; i < data_array.length; i++)
                    _negs[i] = Integer.parseInt(data_array[i]);
        	}
        }

        // </pos>
        if(qualifiedName.equals("pos")) {
        	if(chunk.length() > 0) {
                String data_array[] = chunk.split(",");
                for(int i = 0; i < data_array.length; i++)
                    _poss[i] = Integer.parseInt(data_array[i]);
        	}
        }

        // </cv>
        if(qualifiedName.equals("cv")) {
        	if(chunk.length() > 0) {
                String data_array[] = chunk.split(",");
                for(int i = 0; i < data_array.length; i++)
                    _context[i] = Float.valueOf(data_array[i]).floatValue();
            	_foundCV = true;
        	}
        }

        // </rl>
        if(qualifiedName.equals("rl")) {
            RandomLabel rlabel = new RandomLabel(_word, _meta_data, 
            		_termfrequency, _docfrequency, _negs, _poss, _context);
        	// If no context vector was to be found then the label should be pruned
        	if(!_foundCV)
        		rlabel.prune();
    		_foundCV = false;
            randomIndex.put(_word, rlabel);
            wordsIndexed += _termfrequency;
            if(progressIndicator)
                System.out.print(".");
        }
    }

    public void characters(char[] ch, int start, int length) {
        if(buffer != null)
            buffer.append(ch, start, length);
    }

    public void endDocument() {
        if(progressIndicator)
            System.out.print("\n");
    }

}
