/*
 * RandomIndex.java
 * Copyright (c) 2004, KTH NADA.
 *
 * This file is part of SweSum^2 (see http://swesum.nada.kth.se/),
 * and is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, see
 * http://opensource.org/licenses/gpl-license.php
 *
 * Martin Hassel, 2004-Feb-01
 * http://www.nada.kth.se/~xmartin/
 *
 */

package moj.ri;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import moj.ri.weighting.MangesWS;
import moj.ri.weighting.WeightingScheme;

/**
 * A Random Index is an index that contextually indexes the texts that are fed
 * to it. Not only is it possible to ask the RandomIndex for the term frequency
 * for a specific index term, but also for its semantically closest relatives
 * (based upon co-occurrence statistics and random labels).
 *
 * @author  Martin Hassel
 * @version 2006-Oct-18
 */
public class RandomIndex implements Serializable {
	private static final long serialVersionUID = 6735834893960147189L;
	private int _maxNumThreads = Runtime.getRuntime().availableProcessors();
	private ExecutorService executor = Executors.newFixedThreadPool(_maxNumThreads);
	long _sleepFactor = 50;
	/** The dimensionality of newly created RandomLabels */
    protected int _dimensionality;
    /** Number of non-zeros in newly created RandomLabels */
    protected int _randomDegree;
    /** Seed used in conjunction with the word to generate random labels */
    protected int _seed;
    /** Size of the context window to the left of the focus word */
    protected int _leftWindowSize;
    /** Size of the context window to the right of the focus word */
    protected int _rightWindowSize;
    /** Weighting Scheme applied to the context label when a random label's context is updated */
    protected WeightingScheme _weightingScheme;
    /** The actual RandomIndex */
    protected Map<String,RandomLabel> _randomIndex = new ConcurrentHashMap<String,RandomLabel>();
    /** The vocabulary set, if one is used */
    protected Set<String> _vocabulary = new HashSet<String>();
    /** The number of <i>word instances</i> indexed so far */
    protected long _wordsIndexed = 0;
    /** The number of <i>documents</i> indexed so far */
    protected long _documentsIndexed = 0;
    /** Should indexed words be all lowercase? */
    protected boolean _allLowerCase = true;
    /** Should the random label be unary rather than random? */
    protected boolean _unaryLabels = false;
    /** Should indexing be done on document level rather than word context level? */
    protected boolean _documentLabels = false;
    /** Should RandomLabels be created as new words occur in right context? */
    protected boolean _createAhead = true;
    /** Have labels been removed from the index? */
    protected boolean _purged = false;

    /**
     * Create a new RandomIndex of RandomLabels with the given dimensionality,
     * degree of initial randomness and window size for contextual updates.
     * @param dimensionality The dimensionality all RandomLabels in the RandomIndex
     *        should have, this can not be altered at a later state.
     * @param randomDegree The number of random values all RandomLabels in the
     *        RandomIndex initially should have.
     * @param seed a seed for each label's random generator. This seed, in combination
     *        with the word, makes it very likely that the created random label is
     *        "unique" yet reproducable.
     * @param leftWindowSize The maximum number of words behind the focus word
     *        to include in the context window when updating a label.
     * @param rightWindowSize The maximum number of words in front of the focus
     *        word to include in the context window when updating a label.
     * @param weightingScheme a visiting object defining the weighting of the context
     *        labels to the left and to the right of the focus word.
     */
    public RandomIndex(int dimensionality, int randomDegree, int seed, 
    		int leftWindowSize, int rightWindowSize, WeightingScheme weightingScheme) {
        _dimensionality  = dimensionality;
        _randomDegree   = randomDegree;
        _seed = seed;
        _leftWindowSize  = leftWindowSize;
        _rightWindowSize = rightWindowSize;
        _weightingScheme  = weightingScheme;
    }

    /**
     * Create a new RandomIndex of RandomLabels with a dimensionality of 1800,
     * a degree of initial randomness of 8 and a window size for contextual
     * updates of 3 (i.e. three words look-behind and look-ahead). The seed
     * for randomization is set to 123.
     */
    public RandomIndex() {
        _dimensionality = 1800;
        _randomDegree  = 8;
        _seed = 123;
        _leftWindowSize  = 3;
        _rightWindowSize = 3;
        _weightingScheme  = new MangesWS(); 
    }

    /**
     * Returns a set view of the keys contained in this <code>RandomIndex</code>. 
     * The set is backed by the <code>RandomIndex</code>, so changes to the 
     * <code>RandomIndex</code> are reflected in the set, and vice-versa. 
     * If the <code>RandomIndex</code> is modified while an iteration over the 
     * set is in progress, the results of the iteration are undefined.
     * @return a set view of the keys contained in this <code>RandomIndex</code>.
     */
    public Set<String> keySet() {
    	return _randomIndex.keySet();
    }

    /**
     * Returns a set view of the mappings contained in this <code>RandomIndex</code>. 
     * Each element in the returned set is a <code>Map.Entry</code>. The set is 
     * backed by the <code>RandomIndex</code>, so changes to the 
     * <code>RandomIndex</code> are reflected in the set, and vice-versa. 
     * If the <code>RandomIndex</code> is modified while an iteration over the 
     * set is in progress, the results of the iteration are undefined.
     * @return a set view of the mappings contained in this <code>RandomIndex</code>.
     */
    public Set<Map.Entry<String,RandomLabel>> entrySet() {
    	return _randomIndex.entrySet();
    }

    /**
     * Add an existing RandomLabel to the RandomIndex. Will only succeed if
     * the word (index term) associated with the label does not yet exist in
     * the RandomIndex and the label has the same dimensionality as is set
     * for the RandomIndex.
     * @param label RandomLabel to be added.
     * @return <code>true</code> upon success, else <code>false</code>.
     */
    public boolean addRandomLabel(RandomLabel label) {
        if(label.getDimensionality() != _dimensionality)
        	return false;
        if(_randomIndex.containsKey(label.getWord()))
        	return false;
        _randomIndex.put(label.getWord(),label);
        _wordsIndexed += label.getTermFrequency();
        return true;
    }

    /**
     * Sets the RandomIndex to henceforth use unary labels, incremented from 
     * seed, and returns the previous state. Index terms already in the index 
     * will <i>not</i> be changed.
     * @param unaryLabels boolean value denoting the wish to henceforth use 
     *        unary labels, rather than random, or not.
     * @return the previous state, i.e. either <code>true</code> or <code>false</code>.
     */
    public boolean setUnaryLabels(boolean unaryLabels) {
        boolean oldUnaryLabels = _unaryLabels;
        _unaryLabels = unaryLabels;
        return oldUnaryLabels;
    }

    /**
     * Gets the state of the RandomIndex denoting the wish to henceforth
     * use unary labels, rather than random.
     * @return the current state, i.e. either <code>true</code> or <code>false</code>.
     */
    public boolean getUnaryLabels() {
        return _unaryLabels;
    }

    /**
     * Sets the RandomIndex to henceforth index on document level instead of 
     * the usual word context level. Index terms already in the index 
     * will <i>not</i> be changed.
     * @param documentLabels boolean value denoting the wish to henceforth 
     *        index on document level or not.
     * @return the previous state, i.e. either <code>true</code> or <code>false</code>.
     */
    public boolean setDocumentLabels(boolean documentLabels) {
        boolean oldDocumentLabels = _documentLabels;
        _documentLabels = documentLabels;
        return oldDocumentLabels;
    }

    /**
     * Gets the state of the RandomIndex denoting the wish to henceforth
     * index on document level rather than word context level.
     * @return the current state, i.e. either <code>true</code> or <code>false</code>.
     */
    public boolean getDocumentLabels() {
        return _documentLabels;
    }

    /**
     * Sets the RandomIndex to henceforth automagically convert all words
     * (index terms) to all lower case and returns the previous state. Index terms
     * already in the index will <i>not</i> be changed.
     * @param allLowerCase boolean value denoting the wish to have all index terms henceforth
     *        converted to lower case or not.
     * @return the previous state, i.e. either <code>true</code> or <code>false</code>.
     */
    public boolean setAllLowerCase(boolean allLowerCase) {
        boolean oldAllLowercase = _allLowerCase;
        _allLowerCase = allLowerCase;
        return oldAllLowercase;
    }

    /**
     * Gets the state of the RandomIndex denoting the wish to henceforth
     * automagically convert all words (index terms) to all lower case.
     * @return the current state, i.e. either <code>true</code> or <code>false</code>.
     */
    public boolean getAllLowerCase() {
        return _allLowerCase;
    }

    /**
     * Gets the dimensionality of RandomLabels in the RandomIndex.
     * @return the dimensionality of RandomLabels in the RandomIndex.
     */
    public int getDimensionality() {
        return _dimensionality;
    }

	/**
	 * Gets the random degree of RandomLabels in the RandomIndex.
	 * @return the random degree of RandomLabels in the RandomIndex.
	 */
    public int getRandomDegree() {
		return _randomDegree;
	}

    /**
	 * Gets the seed used to create RandomLabels in the RandomIndex.
	 * @return the seed used to create RandomLabels in the RandomIndex.
	 */
	public int getSeed() {
		return _seed;
	}

	/**
	 * Gets the size of the context window to the left used when updating 
	 * context labels in RandonLabels in the RandomIndex.
	 * @return the size of the context window to the left.
	 */
	public int getLeftWindowSize() {
		return _leftWindowSize;
	}

	/**
	 * Gets the size of the context window to the right used when updating 
	 * context labels in RandonLabels in the RandomIndex.
	 * @return the size of the context window to the right.
	 */
	public int getRightWindowSize() {
		return _rightWindowSize;
	}

	/**
	 * Gets the name (class) of the WeightingScheme used by the RandomIndex 
	 * when updating context labels in RandonLabels in the RandomIndex.
	 * @return the name (class) of the WeightingScheme used by the RandomIndex.
	 */
	public WeightingScheme getWeightingScheme() {
		return _weightingScheme;
	}

	/**
     * Gets the number of words indexed by the Random Index.
     * @return the number of words indexed by the Random Index.
     */
    public long getWordsIndexed() {
        return _wordsIndexed;
    }

    /**
     * Gets the number of documents indexed by the Random Index.
     * @return the number of documents indexed by the Random Index.
     */
    public long getDocumentsIndexed() {
        return _documentsIndexed;
    }

    /**
     * Generates a new random or unary label for a given word and adds it to 
     * the Random Index.
     * @param word index term to create a new Random Label for
     * @return the newly created and to the Random Index added Random Label
     */
    RandomLabel generateLabel(String word) {
    	RandomLabel label;
    	if(_unaryLabels) {
    		// Increment unary position up to, but not beyond, dimensionality
    		int[] unary = {_seed<_dimensionality?_seed++:_dimensionality-1};
    		label = new RandomLabel(word, null, 0, 0, 
    				unary, new int[0], new float[_dimensionality]);
    	} else
        	// If RandomIndex is not in a unary state, create "normal" RandomLabel
    		label = new RandomLabel(word, _dimensionality, _randomDegree, _seed);
        _randomIndex.put(word, label);
    	return label;
    }

    /**
     * The maximum number of concurrent threads that should be reserved for 
     * RI for adding text to the index. 
     * @param maxNumThreads maximum number of threads reserved for RI
     * @return the previously set maximum number of threads reserved for RI
     */
    public int setMaxNumThreads(int maxNumThreads) {
		int oldMaxNumThreads = _maxNumThreads;
		_maxNumThreads = maxNumThreads;
		executor.shutdown();
		executor = Executors.newFixedThreadPool(_maxNumThreads);
		return oldMaxNumThreads;
	}

    /**
     * Returns the maximum number of concurrent threads reserved for RI 
     * when adding text to the index. 
     * @return the previously set maximum number of threads reserved for RI
     */
    public int getMaxNumThreads() {
		return _maxNumThreads;
	}

	/**
	 * The number of milliseconds a thread should sleep when trying to update a 
	 * RandomLabel that is already undergoing updating by another thread.
	 * @param millis the new sleep factor
	 * @return the previously set sleep factor
	 */
	public long setSleepFactor(long millis) {
		long oldSleepFactor = _sleepFactor;
		_sleepFactor = millis;
		return oldSleepFactor;
	}

	/**
	 * Returns the number of milliseconds a thread should sleep when trying to 
	 * update a RandomLabel that is already undergoing updating by another thread.
	 * @return the previously set sleep factor
	 */
	public long getSleepFactor() {
		return _sleepFactor;
	}

    /**
     * Checks if all threads updating the RandomIndex have finished running, only 
     * then the RandomIndex is guaranteed to be up to date. This will never 
     * return <code>true</code> before <code>finalize()</code> has been called.
     * @return <code>true</code> if all threads have finished running, 
     * 			otherwise <code>false</code>
     */
	public boolean finishedUpdating() {
        return executor.isTerminated();
    }

	/**
	 * Shuts down the thread pool, used for scheduling thread (re)usage when 
	 * adding texts to the RandomIndex, as soon as all threads updating the 
	 * RandomIndex have finished running. Should be done e.g. before saving 
	 * the index.
	 */
    public void finalize() {
		executor.shutdown();
	}

    /**
     * Add text to the Random Index and contextually update all words in, and
     * added to, the index. The words updated/added are contextually "coloured"
     * according to the set window size and weighting scheme.
     * @param text a text tokenized on word level where each element in the
     *        list <code>text</code> contains one token.
     * @return number of words (or index terms) added/updated.
     */
    public int addText(String[] text) {
        // If all words are to be lowercased, then lowercase
        if(_allLowerCase)
            for (int current = 0; current < text.length; current++) {
                text[current] = text[current].toLowerCase();
            }

    	AddTextThread thread = new RandomIndex.AddTextThread(text);
    	if(_maxNumThreads <= 1)
        	thread.run();
    	else {
    		if(executor == null || executor.isTerminated())
    			executor = Executors.newFixedThreadPool(_maxNumThreads);
    		executor.submit(thread);
    	}
    	return text.length;
    }

    /**
     * Add text to the Random Index and contextually update all words in, and
     * added to, the index, tokens are separated according to the supplied
     * pattern. The words updated/added are contextually "coloured"
     * according to the set window size and weighting scheme.
     * @param text a text that is to be tokenized according to the supplied
     *        <code>pattern</code>.
     * @param pattern a pattern according to which the text string is to be 
     *        split into tokens.
     * @return number of words (index terms) added/updated.
     */
    public int addText(String text, String pattern) {
        return addText(text.split(pattern));
    }

    /**
     * Add text to the Random Index and contextually update all words in, and
     * added to, the index, tokens are separated by white-space. The words
     * updated/added are contextually "coloured" according to the set window
     * size and weighting scheme.
     * @param text a text that is to be tokenized according to the default
     *        pattern which is <code>"\\s"</code>.
     * @return number of words (index terms) added/updated.
     */
    public int addText(String text) {
        return addText(text, "\\s");
    }

    /**
     * Get the <code>RandomLabel</code> for the given <code>word</code> if it 
     * exists in the <code>RandomIndex</code>. If the <code>RandomIndex</code> 
     * does not contain a corrsponding <code>RandomLabel</code>, return a 
     * zero sized <code>RandomLabel</code>.
     * @param word the word in the <code>RandomIndex</code> that we want the 
     *        corresponding <code>RandomLabel</code> for.
     * @return the <code>RandomLabel<code> corresponding to the given 
     *         <code>word</code> if it exists in the <code>RandomIndex</code>, 
     *         if not a zero sized <code>RandomLabel</code> is returned.
     */
    public RandomLabel getRandomLabel(String word) {
        if(!_randomIndex.containsKey(word)) return new RandomLabel("",0,0,0);
        return _randomIndex.get(word);
    }

    /**
     * String representation of RandomIndex.
     * @return a string containing string representations of RandomLabels
     *         where each label is separated by a newline (the string also
     *         ends in newline).
     */
    public String toString() {
        StringBuffer retstring = new StringBuffer();
        Iterator kit = _randomIndex.keySet().iterator();
        while(kit.hasNext())
            retstring.append(_randomIndex.get(kit.next())+"\n");
        return retstring.toString();
    }

    /**
     * Returns <code>true</code> if this RandomIndex contains no elements.
     * @return <code>true</code> if this RandomIndex contains no elements, 
     *         otherwise false.
     */
    public boolean isEmpty() {
        return _randomIndex.isEmpty();
    }

    /**
     * Returns the number of Random Labes in the Random Index.
     * @return the number of Random Labes in the Random Index.
     */
    public int size() {
        return _randomIndex.size();
    }

    /**
     * Returns <code>true</code> if this RandomIndex contains the given word (index term).
     * @param word The word which existance in the RandomIndex is to be determined.
     * @return <code>true</code> if this set contains the word, otherwise false.
     */
    public boolean contains(String word) {
        return _randomIndex.containsKey(word);
    }

    /**
     * Returns <code>true</code> if this RandomIndex contains the given RandomLabel.
     * @param label The RandomLabel which existance in the RandomIndex is to be determined.
     * @return <code>true</code> if this set contains the RandomIndex, otherwise false.
     */
    public boolean contains(RandomLabel label) {
        return (label == getRandomLabel(label.getWord()));
    }

    /**
	 * Prunes the RandomLabels associated to the given words by setting their 
	 * context vectors to <code>null</code>. This can be used to lower memory 
	 * constraints and save file size if there are a lot of tokens that are 
	 * only interesting as context to other tokens, and not as actual index 
	 * terms themselves.
     * @param words a <code>string<code> vector of words whose RandomLabels 
     *        are to be pruned.
     */
    public void pruneIndex(String[] words) {
    	for(int i=0; i < words.length; i++)
    		getRandomLabel(words[i]).prune();
    }

    /**
     * Prunes RandomLabels with a term frequency lower than <code>minTF</code> 
     * or higher than <code>maxTF</code> by setting their context vectors to 
     * <code>null</code>. This can be used to lower memory constraints and save 
     * file size if there are a lot of tokens that are only interesting as 
     * context to other tokens, and not as actual index terms themselves.
     * @param minTF lower frequency theshold
     * @param maxTF higher frequency treshold
     */
    public void pruneIndex(long minTF, long maxTF) {
    	for(Iterator iterator = _randomIndex.keySet().iterator(); iterator.hasNext();) {
        	RandomLabel label = _randomIndex.get(iterator.next()); 
        	if(label.getTermFrequency() < minTF)
        		label.prune();
        	else if(label.getTermFrequency() > maxTF)
        		label.prune();
        }
    }

    /**
	 * Removes the <code>RandomLabels</code> associated to tokens previously pruned.
	 * @return the number of <code>RandomLabels</code> removed
     */
    public int purgeIndex() {
    	int removed = 0;
    	for(Iterator iterator = _randomIndex.keySet().iterator(); iterator.hasNext();) {
    		RandomLabel label = _randomIndex.get(iterator.next());
            if(label.getContext() == null) {
                _randomIndex.remove(label.getWord());
                removed++;
            }
		}
    	if(removed != 0)
            _purged = true;
        return removed;
    }

    /**
	 * Removes the <code>RandomLabels</code> associated to the given words from the 
	 * <code>RandomIndex</code>.
     * @param words a <code>string<code> vector of words whose RandomLabels 
     *        are to be removed.
	 * @return the number of <code>RandomLabels</code> removed
     */
    public int purgeIndex(String[] words) {
    	int removed = 0;
    	for(int i=0; i < words.length; i++) {
    		_randomIndex.remove(words[i]);
    		removed++;
    	}
        if(removed != 0)
            _purged = true;
        return removed;
    }

    /**
     * Removes <code>RandomLabels</code> with a term frequency lower than <code>minTF</code> 
     * or higher than <code>maxTF</code> from the <code>RandomIndex</code>.
     * @param minTF lower frequency theshold
     * @param maxTF higher frequency treshold
	 * @return the number of <code>RandomLabels</code> removed
     */
    public int purgeIndex(long minTF, long maxTF) {
    	int removed = 0;
    	for(Iterator iterator = _randomIndex.keySet().iterator(); iterator.hasNext();) {
        	RandomLabel label = _randomIndex.get(iterator.next()); 
        	if(label.getTermFrequency() < minTF) {
        		_randomIndex.remove(label.getWord());
        		removed++;
        	} else if(label.getTermFrequency() > maxTF) {
        		_randomIndex.remove(label.getWord());
        		removed++;
        	}
        }
        if(removed != 0)
            _purged = true;
        return removed;
    }

    /**
     * Note: Updating the RandomIndex after labels have been removed from the 
     * index may result in new labels being generated for tokens that have 
     * already been encountered. This will lead to incorrectly updated context 
     * vectors and unreliable cooccurrence data.
     * @return <code>true</code> if labels have been removed from the index, 
     *         if not <code>false</code>
     */
    public boolean isPurged() {
    	return _purged;
    }
    
    /** 
     * Adds words (index terms) to the restricted vocabulary. Words <i>not</i> 
     * in the restricted vocabulary will have <code>RandomLabel</code>s with 
     * pruned context vectors (i.e. contextvector = <b><code>null</code></b>).
     * @param words <code>String</code> array with words to be added
     */
    public void addVocabulary(String[] words) {
    	for(int i=0; i < words.length; i++)
			_vocabulary.add(words[i]);
    }

    /** 
     * Adds words (index terms) to the restricted vocabulary. Words <i>not</i> 
     * in the restricted vocabulary will have <code>RandomLabel</code>s with 
     * pruned context vectors (i.e. contextvector = <b><code>null</code></b>).
     * @param vocabulary <code>Set</code> with words to be added
     */
    public int addVocabulary(Set<String> vocabulary) {
    	int added = 0;
    	for(Iterator iterator = vocabulary.iterator(); iterator.hasNext(); added++) {
			_vocabulary.add((String)iterator.next());
		}
    	return added;
    }

    /**
     * Sets the restricted vocabulary to the supplied <code>Set</code> and 
     * returns the old vocabulary set.<br />
     * In order to reset the vocabulary you can supply an empty <code>Set</code>, 
     * the <code>RandomLabel</code>s with pruned context vectors will however 
     * stay pruned (i.e. contextvector = <b><code>null</code></b>).
     * 
     * @param vocabulary <code>Set</code> of words (index terms) denoting the 
     *        desired restricted vocabulary
     * @return previous vocabulary <code>Set</code> (possibly empty) 
     */
    public Set<String> setVocabulary(Set<String> vocabulary) {
    	Set<String> tempvoc = _vocabulary;
		_vocabulary = vocabulary;
		return tempvoc;
    }

    /**
     * Gets the <code>Set</code> denoting the current restricted vocabulary.
     * @return current vocabulary <code>Set</code> (possibly empty) 
     */
    public Set<String> getVocabulary() {
		return _vocabulary;
    }

    private class AddTextThread extends Thread {
    	private String[] text;

    	AddTextThread(String[] txt) {
    		text = txt;
		}

        public void run() {
            Set<String> uniqueWords = new HashSet<String>();
            // Construct empty RandomLabel to ensure all zeros in context
            RandomLabel padLabel = new RandomLabel();
            RandomLabel[] left = new RandomLabel[1];
            RandomLabel[] right = new RandomLabel[0];

            // If indexing on document level, create context denoting current document
            if(_documentLabels) {
            	RandomLabel docLabel;
            	if(_unaryLabels) {
            		// Increment unary position up to, but not beyond, dimensionality
            		int[] unary = {_documentsIndexed<_dimensionality?
            				(int)_documentsIndexed:_dimensionality-1};
            		docLabel = new RandomLabel("doc"+_documentsIndexed, null, 0, 0, 
            				unary, new int[0], new float[_dimensionality]);
            	} else {
                	docLabel = new RandomLabel("doc"+_documentsIndexed, 
                			_dimensionality, _randomDegree, _seed);
            	}
                left[0] = docLabel;
            }

            // Traverse all words in array text and add/update them to/in the RandomIndex
            for(int current = 0; current < text.length; current++) {
                String word = text[current];
                uniqueWords.add(word);

                if(!_documentLabels) {
                    // Create left and right context window (arrays of RandomLabels)
                    left = new RandomLabel[_leftWindowSize];
                    for(int distance = 0; distance < _leftWindowSize; distance++) {
                        // Do not go beyond left array boundary
                        if((current-distance-1) < 0) {
                            left[distance] = padLabel;
                        } else {
                        	if(_randomIndex.containsKey(text[current-distance-1])) {
                        		 left[distance] = _randomIndex.get(text[current-distance-1]);
                        	} else {
                        		 // A context word to the left is not yet in the index???
                        		 left[distance] = generateLabel(text[current-distance-1]);
                        	}
                        }
                    }
                    right = new RandomLabel[_rightWindowSize];
                    for(int distance = 0; distance < _rightWindowSize; distance++) {
                        // Do not go beyond right array boundary
                        if((current+distance+1) > text.length-1) {
                            right[distance] = padLabel;
                        } else {
                            if(_randomIndex.containsKey(text[current+distance+1]))
                                right[distance] = _randomIndex.get(text[current+distance+1]);
                            else
                                // A context word to the right is not yet in the index
                            	if(_createAhead)
                            		right[distance] = generateLabel(text[current+distance+1]);
                            	else
                            		right[distance] = padLabel;
                        }
                    }
                }

            	RandomLabel wordLabel;
                // If the word already exists in the index we just collect it from there
                if(_randomIndex.containsKey(word))
                    wordLabel = _randomIndex.get(word);
                // If not, which is true for each first occurrence, we need to create it
                else {
                	wordLabel = generateLabel(word);
    	            _randomIndex.put(word, wordLabel);
                }

    			// If the word is in the vocabulary to be indexed or if no vocabulary is set
    			if(_vocabulary.contains(word) || _vocabulary.isEmpty()) {
    				// If the RL is not in a valid state, ie another thread is updating it
    				while(!wordLabel.validState()) {
    					try {
    						Thread.sleep(_sleepFactor);
    					} catch(Exception ex) {
    						System.err.println(Thread.currentThread().getName()+"Sleep failed");
    						ex.printStackTrace();
    					}
    				}
    				// Update the RandomLabel's context vector
    	            wordLabel.updateContext(left, right, _weightingScheme);
    			} else {
    				_randomIndex.get(word).incrementTermFrequency();
    				_randomIndex.get(word).prune();
    			}
                _wordsIndexed++;
            }

            Iterator uwit = uniqueWords.iterator();
            while(uwit.hasNext()) {
                String word = (String)uwit.next();
                RandomLabel label = _randomIndex.get(word);
                label.incrementDocumentFrequency();
                _randomIndex.put(word,label);
            }
            _documentsIndexed++;
        }
    }
}
